/*
 * stream.cpp - handles a client stream
 * Copyright (C) 2003  Justin Karneges
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */

/*
  Notes:
    - For Non-SASL auth (JEP-0078), username and resource fields are required.
*/

#include"xmpp.h"

#include<qtextstream.h>
#include<qguardedptr.h>
#include<qtimer.h>
#include<qca.h>
#include"bytestream.h"
#include"base64.h"
#include"hash.h"
#include"simplesasl.h"

#ifdef XMPP_TEST
#include"td.h"
#endif

using namespace XMPP;

#define NS_XML     "http://www.w3.org/XML/1998/namespace"
#define NS_STREAMS "http://etherx.jabber.org/streams"
#define NS_CLIENT  "jabber:client"
#define NS_TLS     "urn:ietf:params:xml:ns:xmpp-tls"
#define NS_SASL    "urn:ietf:params:xml:ns:xmpp-sasl"
#define NS_SESSION "urn:ietf:params:xml:ns:xmpp-session"
#define NS_STANZAS "urn:ietf:params:xml:ns:xmpp-stanzas"

static Debug *debug_ptr = 0;
void XMPP::setDebug(Debug *p)
{
	debug_ptr = p;
}

//----------------------------------------------------------------------------
// Stanza
//----------------------------------------------------------------------------
Stanza::Error::Error(int _type, int _condition, const QString &_text, const QDomElement &_appSpec)
{
	type = _type;
	condition = _condition;
	text = _text;
	appSpec = _appSpec;
}

class Stanza::Private
{
public:
	struct ErrorTypeEntry
	{
		const char *str;
		int type;
	};
	static ErrorTypeEntry errorTypeTable[];

	struct ErrorCondEntry
	{
		const char *str;
		int cond;
	};
	static ErrorCondEntry errorCondTable[];

	static int stringToKind(const QString &s)
	{
		if(s == "message")
			return Message;
		else if(s == "presence")
			return Presence;
		else if(s == "iq")
			return IQ;
		else
			return -1;
	}

	static QString kindToString(Kind k)
	{
		if(k == Message)
			return "message";
		else if(k == Presence)
			return "presence";
		else
			return "iq";
	}

	static int stringToErrorType(const QString &s)
	{
		for(int n = 0; errorTypeTable[n].str; ++n) {
			if(s == errorTypeTable[n].str)
				return errorTypeTable[n].type;
		}
		return -1;
	}

	static QString errorTypeToString(int x)
	{
		for(int n = 0; errorTypeTable[n].str; ++n) {
			if(x == errorTypeTable[n].type)
				return errorTypeTable[n].str;
		}
		return QString();
	}

	static int stringToErrorCond(const QString &s)
	{
		for(int n = 0; errorCondTable[n].str; ++n) {
			if(s == errorCondTable[n].str)
				return errorCondTable[n].cond;
		}
		return -1;
	}

	static QString errorCondToString(int x)
	{
		for(int n = 0; errorCondTable[n].str; ++n) {
			if(x == errorCondTable[n].cond)
				return errorCondTable[n].str;
		}
		return QString();
	}

	Stream *s;
	QDomElement e;
};

Stanza::Private::ErrorTypeEntry Stanza::Private::errorTypeTable[] =
{
	{ "cancel",   Cancel },
	{ "continue", Continue },
	{ "modify",   Modify },
	{ "auth",     Auth },
	{ "wait",     Wait },
	{ 0, 0 },
};

Stanza::Private::ErrorCondEntry Stanza::Private::errorCondTable[] =
{
	{ "bad-request",             BadRequest },
	{ "conflict",                Conflict },
	{ "feature-not-implemented", FeatureNotImplemented },
	{ "forbidden",               Forbidden },
	{ "internal-server-error",   InternalServerError },
	{ "item-not-found",          ItemNotFound },
	{ "jid-malformed",           JidMalformed },
	{ "not-allowed",             NotAllowed },
	{ "payment-required",        PaymentRequired },
	{ "recipient-unavailable",   RecipientUnavailable },
	{ "registration-required",   RegistrationRequired },
	{ "remote-server-not-found", ServerNotFound },
	{ "remote-server-timeout",   ServerTimeout },
	{ "resource-constraint",     ResourceConstraint },
	{ "service-unavailable",     ServiceUnavailable },
	{ "subscription-required",   SubscriptionRequired },
	{ "undefined-condition",     UndefinedCondition },
	{ "unexpected-request",      UnexpectedRequest },
	{ 0, 0 },
};

Stanza::Stanza()
{
	d = 0;
}

Stanza::Stanza(Stream *s, Kind k, const Jid &to, const QString &type, const QString &id)
{
	d = new Private;

	Kind kind;
	if(k == Message || k == Presence || k == IQ)
		kind = k;
	else
		kind = Message;

	d->s = s;
	d->e = d->s->doc().createElementNS(s->baseNS(), Private::kindToString(kind));
	if(to.isValid())
		setTo(to);
	if(!type.isEmpty())
		setType(type);
	if(!id.isEmpty())
		setId(id);
}

Stanza::Stanza(Stream *s, const QDomElement &e)
{
	d = 0;
	if(e.namespaceURI() != s->baseNS())
		return;
	int x = Private::stringToKind(e.tagName());
	if(x == -1)
		return;
	d = new Private;
	d->s = s;
	d->e = e;
}

Stanza::Stanza(const Stanza &from)
{
	d = 0;
	*this = from;
}

Stanza & Stanza::operator=(const Stanza &from)
{
	delete d;
	d = 0;
	if(from.d)
		d = new Private(*from.d);
	return *this;
}

Stanza::~Stanza()
{
	delete d;
}

bool Stanza::isNull() const
{
	return (d ? false: true);
}

QDomElement Stanza::element() const
{
	return d->e;
}

QString Stanza::toString() const
{
	return Stream::xmlToString(d->e);
}

QDomDocument & Stanza::doc() const
{
	return d->s->doc();
}

QString Stanza::baseNS() const
{
	return d->s->baseNS();
}

QDomElement Stanza::createElement(const QString &ns, const QString &tagName)
{
	return d->s->doc().createElementNS(ns, tagName);
}

QDomElement Stanza::createTextElement(const QString &ns, const QString &tagName, const QString &text)
{
	QDomElement e = d->s->doc().createElementNS(ns, tagName);
	e.appendChild(d->s->doc().createTextNode(text));
	return e;
}

void Stanza::appendChild(const QDomElement &e)
{
	d->e.appendChild(e);
}

Stanza::Kind Stanza::kind() const
{
	return (Kind)Private::stringToKind(d->e.tagName());
}

void Stanza::setKind(Kind k)
{
	d->e.setTagName(Private::kindToString(k));
}

Jid Stanza::to() const
{
	return Jid(d->e.attribute("to"));
}

Jid Stanza::from() const
{
	return Jid(d->e.attribute("from"));
}

QString Stanza::id() const
{
	return d->e.attribute("id");
}

QString Stanza::type() const
{
	return d->e.attribute("type");
}

QString Stanza::lang() const
{
	return d->e.attributeNS(NS_XML, "lang", QString());
}

void Stanza::setTo(const Jid &j)
{
	d->e.setAttribute("to", j.full());
}

void Stanza::setFrom(const Jid &j)
{
	d->e.setAttribute("from", j.full());
}

void Stanza::setId(const QString &id)
{
	d->e.setAttribute("id", id);
}

void Stanza::setType(const QString &type)
{
	d->e.setAttribute("type", type);
}

void Stanza::setLang(const QString &lang)
{
	d->e.setAttribute("xml:lang", lang);
}

Stanza::Error Stanza::error() const
{
	Error err;
	QDomElement e = d->e.elementsByTagNameNS(d->s->baseNS(), "error").item(0).toElement();
	if(e.isNull())
		return err;

	// type
	int x = Private::stringToErrorType(e.attribute("type"));
	if(x != -1)
		err.type = x;

	// condition
	QDomElement t = e.firstChild().toElement();
	if(!t.isNull() && t.namespaceURI() == NS_STANZAS) {
		x = Private::stringToErrorCond(t.tagName());
		if(x != -1)
			err.condition = x;
	}

	// text
	t = e.elementsByTagNameNS(NS_STANZAS, "text").item(0).toElement();
	if(!t.isNull())
		err.text = t.text();
	else
		err.text = e.text();

	// appspec: find first non-standard namespaced element
	QDomNodeList nl = e.childNodes();
	for(uint n = 0; n < nl.count(); ++n) {
		QDomNode i = nl.item(n);
		if(i.isElement() && i.namespaceURI() != NS_STANZAS) {
			err.appSpec = i.toElement();
			break;
		}
	}
	return err;
}

void Stanza::setError(const Error &err)
{
	// create the element if necessary
	QDomElement errElem = d->e.elementsByTagNameNS(d->s->baseNS(), "error").item(0).toElement();
	if(errElem.isNull()) {
		errElem = d->e.ownerDocument().createElementNS(d->s->baseNS(), "error");
		d->e.appendChild(errElem);
	}

	// error type/condition
	if(d->s->old()) {
		errElem.setAttribute("code", QString::number(err.condition));
	}
	else {
		QString stype = Private::errorTypeToString(err.type);
		if(stype.isEmpty())
			return;
		QString scond = Private::errorCondToString(err.condition);
		if(scond.isEmpty())
			return;

		errElem.setAttribute("type", stype);
		errElem.appendChild(d->e.ownerDocument().createElementNS(d->s->baseNS(), scond));
	}

	// text
	if(d->s->old()) {
		errElem.appendChild(d->e.ownerDocument().createTextNode(err.text));
	}
	else {
		QDomElement te = d->e.ownerDocument().createElementNS(d->s->baseNS(), "text");
		te.appendChild(d->e.ownerDocument().createTextNode(err.text));
		errElem.appendChild(te);
	}

	// application specific
	errElem.appendChild(err.appSpec);
}

void Stanza::clearError()
{
	QDomElement errElem = d->e.elementsByTagNameNS(d->s->baseNS(), "error").item(0).toElement();
	if(!errElem.isNull())
		d->e.removeChild(errElem);
}

//----------------------------------------------------------------------------
// Stream
//----------------------------------------------------------------------------

// This function removes namespace information from various nodes for
// display purposes only (the element is pretty much useless for processing
// after this).  We do this because QXml is a bit overzealous about outputting
// redundant namespaces.
static QDomElement stripExtraNS(const QDomElement &e)
{
	// find closest parent with a namespace
	QDomNode par = e.parentNode();
	while(!par.isNull() && par.namespaceURI().isNull())
		par = par.parentNode();
	bool noShowNS = false;
	if(!par.isNull() && par.namespaceURI() == e.namespaceURI())
		noShowNS = true;

	QDomElement i;
	uint x;
	if(noShowNS)
		i = e.ownerDocument().createElement(e.tagName());
	else
		i = e.ownerDocument().createElementNS(e.namespaceURI(), e.tagName());

	// copy attributes
	QDomNamedNodeMap al = e.attributes();
	for(x = 0; x < al.count(); ++x) {
		QDomAttr a = al.item(x).cloneNode().toAttr();

		// don't show xml namespace
		if(a.namespaceURI() == NS_XML)
			i.setAttribute(QString("xml:") + a.name(), a.value());
		else
			i.setAttributeNodeNS(a);
	}

	// copy children
	QDomNodeList nl = e.childNodes();
	for(x = 0; x < nl.count(); ++x) {
		QDomNode n = nl.item(x);
		if(n.isElement())
			i.appendChild(stripExtraNS(n.toElement()));
		else
			i.appendChild(n.cloneNode());
	}
	return i;
}

Stream::Stream(QObject *parent)
:QObject(parent)
{
}

Stream::~Stream()
{
}

Stanza Stream::createStanza(Stanza::Kind k, const Jid &to, const QString &type, const QString &id)
{
	return Stanza(this, k, to, type, id);
}

Stanza Stream::createStanza(const QDomElement &e)
{
	return Stanza(this, e);
}

void Stream::createStreamXmlTags(const QString &ns, const QString &to, const QString &ver, QString *xmlHeader, QString *streamOpen, QString *streamClose)
{
	// use DOM to create the opening and closing stream tags
	QDomDocument doc;
	QDomElement e = doc.createElementNS(NS_STREAMS, "stream:stream");
	e.setAttribute("xmlns", ns); // seems this is the only way to get both namespaces in here
	e.setAttribute("to", to);
	if(!ver.isEmpty())
		e.setAttribute("version", ver);
	QDomElement dummy = doc.createElement("dummy");
	e.appendChild(dummy);
	QString str;
	QTextStream ts(&str, IO_WriteOnly);
	e.save(ts, 0);
	int n = str.find('<');
	int n2 = str.find('>', n);
	++n2;
	*streamOpen = str.mid(n, n2-n);
	n2 = str.findRev('>');
	n = str.findRev('<');
	++n2;
	*streamClose = str.mid(n, n2-n);
	*xmlHeader = "<?xml version=\"1.0\"?>";
}

int Stream::writeString(const QString &s)
{
	QCString cs = s.utf8();
	QByteArray a(cs.length());
	memcpy(a.data(), cs.data(), a.size());
	writeData(a);
	return a.size();
}

int Stream::writeElement(const QDomElement &e, bool clip)
{
	if(e.isNull())
		return 0;
	QString out = xmlToString(e, clip);
	return writeString(out);
}

QString Stream::xmlToString(const QDomElement &e, bool clip)
{
	QDomElement i = e.cloneNode().toElement();
	QDomElement fake = e.ownerDocument().createElementNS("jabber:client", "fake");
	fake.appendChild(i);
	fake = stripExtraNS(fake);
	QString out;
	{
		QTextStream ts(&out, IO_WriteOnly);
		fake.firstChild().save(ts, 0);
	}
	// 'clip' means to remove any unwanted (and unneeded) characters, such as a trailing newline
	if(clip) {
		int n = out.findRev('>');
		out.truncate(n+1);
	}
	return out;
}

//----------------------------------------------------------------------------
// ClientStream
//----------------------------------------------------------------------------
enum {
	Idle,
	Connecting,
	NeedStream,
	WaitVersion,
	NeedFeatures,
	WaitTLS,
	NeedTLSProceed,
	NeedSASLChallenge,
	NeedSASLStep,
	NeedParams,
	NeedAuthGetResponse,
	NeedAuthSetResponse,
	Active,
	Closing
};

class StreamFeatures
{
public:
	StreamFeatures()
	{
		tls_supported = false;
		sasl_supported = false;
		tls_required = false;
	}

	bool tls_supported, sasl_supported;
	bool tls_required;
	QStringList sasl_mechs;
};

class LayerTracker
{
public:
	struct Item
	{
		int plain;
		int encoded;
	};

	LayerTracker()
	{
		p = 0;
	}

	void reset()
	{
		p = 0;
		list.clear();
	}

	void addPlain(int plain)
	{
		p += plain;
	}

	void specifyEncoded(int encoded, int plain)
	{
		// can't specify more bytes than we have
		if(plain > p)
			plain = p;
		p -= plain;
		Item i;
		i.plain = plain;
		i.encoded = encoded;
		list += i;
	}

	int finished(int encoded)
	{
		int plain = 0;
		for(QValueList<Item>::Iterator it = list.begin(); it != list.end();) {
			Item &i = *it;

			// not enough?
			if(encoded < i.encoded) {
				i.encoded -= encoded;
				break;
			}

			encoded -= i.encoded;
			plain += i.plain;
			it = list.remove(it);
		}
		return plain;
	}

	int p;
	QValueList<Item> list;
};

class TrackItem
{
public:
	enum Type { Stanza, Ping, Close, Raw };
	TrackItem()
	{
		type = -1;
		size = 0;
	}

	int type;
	int size;
};

class ClientStream::Private
{
public:
	enum SASLCond {
		Aborted,
		BadProtocol,
		EncryptionRequired,
		InvalidAuthzid,
		InvalidMech,
		InvalidRealm,
		MechTooWeak,
		NotAuthorized,
		TemporaryAuthFailure
	};
	enum StreamCond
	{
		BadFormat,
		BadNamespacePrefix,
		Conflict,
		ConnectionTimeout,
		HostGone,
		HostUnknown,
		ImproperAddressing,
		InternalServerError,
		InvalidId,
		InvalidNamespace,
		InvalidXml,
		NonmatchingHosts,
		StreamNotAuthorized,
		PolicyViolation,
		RemoteConnectionFailed,
		ResourceConstraint,
		RestrictedXml,
		SeeOtherHost,
		SystemShutdown,
		UndefinedCondition,
		UnsupportedEncoding,
		UnsupportedStanzaType,
		UnsupportedVersion,
		XmlNotWellFormed
	};

	struct SASLCondEntry
	{
		const char *str;
		int cond;
	};
	static SASLCondEntry saslCondTable[];

	struct StreamCondEntry
	{
		const char *str;
		int cond;
	};
	static StreamCondEntry streamCondTable[];

	Private()
	{
		conn = 0;
		bs = 0;
		tls = 0;
		sasl = 0;
		in.setAutoDelete(true);

		reset();
	}

	void reset()
	{
		sasl_authed = false;
		tls_started = false;
		tls_warned = false;
		allowPlain = false;
		mutualAuth = false;
		minimumSSF = 0;
		maximumSSF = 0;
		oldOnly = false;
		haveLocalAddr = false;
		state = Idle;
		layerTLS.reset();
		layerSASL.reset();
		in.clear();
	}

	static int stringToSASLCond(const QString &s)
	{
		for(int n = 0; saslCondTable[n].str; ++n) {
			if(s == saslCondTable[n].str)
				return saslCondTable[n].cond;
		}
		return -1;
	}

	static int stringToStreamCond(const QString &s)
	{
		for(int n = 0; streamCondTable[n].str; ++n) {
			if(s == streamCondTable[n].str)
				return streamCondTable[n].cond;
		}
		return -1;
	}

	static int getOldErrorCode(const QDomElement &e)
	{
		QDomElement err = e.elementsByTagNameNS(NS_CLIENT, "error").item(0).toElement();
		if(err.isNull() || !err.hasAttribute("code"))
			return -1;
		return err.attribute("code").toInt();
	}

	static int authErrorOldToNew(int x)
	{
		if(x == 401) // not authorized
			return ClientStream::NotAuthorized;
		else if(x == 409) // conflict
			return ClientStream::GenericAuthError;
		else if(x == 406) // not acceptable (this should NOT happen)
			return ClientStream::GenericAuthError;
		return -1;
	}

	QString server;
	Connector *conn;
	ByteStream *bs;
	Parser xml;
	TLSHandler *tls;
	QCA::SASL *sasl;
	QString sasl_mech;
	int sasl_ssf;
	bool sasl_authed, tls_started, tls_warned;
	bool allowPlain, mutualAuth;
	bool old;
	int errCond;
	QString errText;
	QDomElement errAppSpec;
	QDomDocument doc;
	QString streamOpen, streamClose;
	QString id, from;
	int state;
	StreamFeatures features;
	Jid jid;
	QByteArray spare;
	bool digest;
	QString pass;
	int minimumSSF, maximumSSF;
	bool haveLocalAddr;
	QHostAddress localAddr;
	Q_UINT16 localPort;
	QValueList<TrackItem> trackQueue;
	QPtrList<Stanza> in;
	bool oldOnly, doAuth;

	LayerTracker layerTLS, layerSASL;

	QTimer noopTimer;
	int noop_time;
};

ClientStream::Private::SASLCondEntry ClientStream::Private::saslCondTable[] =
{
	{ "aborted",                Aborted },
	{ "bad-protocol",           BadProtocol },
	{ "encryption-required",    EncryptionRequired },
	{ "invalid-authzid",        InvalidAuthzid },
	{ "invalid-mechanism",      InvalidMech },
	{ "invalid-realm",          InvalidRealm },
	{ "mechanism-too-weak",     MechTooWeak },
	{ "not-authorized",         NotAuthorized },
	{ "temporary-auth-failure", TemporaryAuthFailure },
	{ 0, 0 },
};

ClientStream::Private::StreamCondEntry ClientStream::Private::streamCondTable[] =
{
	{ "bad-format",               BadFormat },
	{ "bad-namespace-prefix",     BadNamespacePrefix },
	{ "conflict",                 Conflict },
	{ "connection-timeout",       ConnectionTimeout },
	{ "host-gone",                HostGone },
	{ "host-unknown",             HostUnknown },
	{ "improper-addressing",      ImproperAddressing },
	{ "internal-server-error",    InternalServerError },
	{ "invalid-id",               InvalidId },
	{ "invalid-namespace",        InvalidNamespace },
	{ "invalid-xml",              InvalidXml },
	{ "nonmatching-hosts",        NonmatchingHosts },
	{ "not-authorized",           StreamNotAuthorized },
	{ "policy-violation",         PolicyViolation },
	{ "remote-connection-failed", RemoteConnectionFailed },
	{ "resource-constraint",      ResourceConstraint },
	{ "restricted-xml",           RestrictedXml },
	{ "see-other-host",           SeeOtherHost },
	{ "system-shutdown",          SystemShutdown },
	{ "undefined-condition",      UndefinedCondition },
	{ "unsupported-encoding",     UnsupportedEncoding },
	{ "unsupported-stanza-type",  UnsupportedStanzaType },
	{ "unsupported-version",      UnsupportedVersion },
	{ "xml-not-well-formed",      XmlNotWellFormed },
	{ 0, 0 },
};

ClientStream::ClientStream(Connector *conn, TLSHandler *tls, QObject *parent)
:Stream(parent)
{
	d = new Private;
	d->conn = conn;
	connect(d->conn, SIGNAL(connected()), SLOT(cr_connected()));
	connect(d->conn, SIGNAL(error()), SLOT(cr_error()));

	d->noop_time = 0;
	connect(&d->noopTimer, SIGNAL(timeout()), SLOT(doNoop()));

	d->tls = tls;
	if(d->tls) {
		connect(d->tls, SIGNAL(success()), SLOT(tls_success()));
		connect(d->tls, SIGNAL(fail()), SLOT(tls_fail()));
		connect(d->tls, SIGNAL(closed()), SLOT(tls_closed()));
		connect(d->tls, SIGNAL(readyRead(const QByteArray &)), SLOT(tls_readyRead(const QByteArray &)));
		connect(d->tls, SIGNAL(readyReadOutgoing(const QByteArray &, int)), SLOT(tls_readyReadOutgoing(const QByteArray &, int)));
	}
}

ClientStream::~ClientStream()
{
	reset();
	delete d;
}

void ClientStream::connectToServer(const Jid &jid, bool auth)
{
	reset();
	d->state = Connecting;
	d->jid = jid;
	d->server = d->jid.domain();
	d->doAuth = auth;

	d->conn->connectToServer(d->server);
}

void ClientStream::continueAfterWarning()
{
	if(d->state == WaitVersion) {
		do_afterStream();
	}
	else if(d->state == WaitTLS) {
		if(d->old)
			do_afterStream();
		else
			do_handleFeatures();
	}
}

bool ClientStream::isActive() const
{
	return (d->state != Idle) ? true: false;
}

bool ClientStream::isAuthenticated() const
{
	return (d->state == Active) ? true: false;
}

void ClientStream::setUsername(const QString &s)
{
	if(d->sasl)
		d->sasl->setUsername(s);
}

void ClientStream::setPassword(const QString &s)
{
	if(d->old)
		d->pass = s;
	else {
		if(d->sasl)
			d->sasl->setPassword(s);
	}
}

void ClientStream::setRealm(const QString &s)
{
	if(d->sasl)
		d->sasl->setRealm(s);
}

void ClientStream::continueAfterParams()
{
	if(d->old) {
		do_authset();
	}
	else {
		if(d->sasl && d->state == NeedParams)
			d->sasl->continueAfterParams();
	}
}

void ClientStream::setNoopTime(int mills)
{
	d->noop_time = mills;

	if(d->state != Active)
		return;

	if(d->noop_time == 0) {
		d->noopTimer.stop();
		return;
	}
	d->noopTimer.start(d->noop_time);
}

QString ClientStream::saslMechanism() const
{
	return d->sasl_mech;
}

int ClientStream::saslSSF() const
{
	return d->sasl_ssf;
}

void ClientStream::setSASLMechanism(const QString &s)
{
	d->sasl_mech = s;
}

void ClientStream::setLocalAddr(const QHostAddress &addr, Q_UINT16 port)
{
	d->haveLocalAddr = true;
	d->localAddr = addr;
	d->localPort = port;
}

int ClientStream::errorCondition() const
{
	return d->errCond;
}

QString ClientStream::errorText() const
{
	return d->errText;
}

QDomElement ClientStream::errorAppSpec() const
{
	return d->errAppSpec;
}

bool ClientStream::old() const
{
	return d->old;
}

void ClientStream::close()
{
	if(d->state == Active) {
		d->state = Closing;
#ifdef XMPP_TEST
		TD::outgoingTag(d->streamClose);
#endif
		TrackItem i;
		i.type = TrackItem::Close;
		i.size = writeString(d->streamClose);
		d->trackQueue += i;
	}
	else if(d->state != Idle && d->state != Closing) {
		reset();
	}
}

QDomDocument & ClientStream::doc() const
{
	return d->doc;
}

QString ClientStream::baseNS() const
{
	return NS_CLIENT;
}

void ClientStream::setAllowPlain(bool b)
{
	d->allowPlain = b;
}

void ClientStream::setRequireMutualAuth(bool b)
{
	d->mutualAuth = b;
}

void ClientStream::setSSFRange(int low, int high)
{
	d->minimumSSF = low;
	d->maximumSSF = high;
}

void ClientStream::setOldOnly(bool b)
{
	d->oldOnly = b;
}

bool ClientStream::stanzaAvailable() const
{
	return (!d->in.isEmpty());
}

Stanza ClientStream::read()
{
	if(d->in.isEmpty())
		return Stanza();
	else {
		Stanza *sp = d->in.getFirst();
		Stanza s = *sp;
		d->in.removeRef(sp);
		return s;
	}
}

void ClientStream::write(const Stanza &s)
{
	if(d->state == Active) {
		TrackItem i;
		i.type = TrackItem::Stanza;
		i.size = writeElement(s.element());
		d->trackQueue += i;
	}
}

void ClientStream::writeDirect(const QString &s)
{
#ifdef XMPP_TEST
	TD::outgoingTag(s);
#endif
	TrackItem i;
	i.type = TrackItem::Raw;
	i.size = writeString(s);
	d->trackQueue += i;
}

void ClientStream::reset(bool)
{
	if(d->state == Idle)
		return;

	d->reset();
	d->noopTimer.stop();

	// reset sasl
	delete d->sasl;
	d->sasl = 0;

	// reset tls
	if(d->tls)
		d->tls->reset();

	// reset connector
	if(d->bs) {
		d->bs->close();
		d->bs = 0;
	}
	d->conn->done();

	// parser
	d->xml.reset();

	d->spare.resize(0);
	d->sasl_mech = QString();
	d->trackQueue.clear();
}

void ClientStream::cr_connected()
{
	d->bs = d->conn->stream();
	connect(d->bs, SIGNAL(connectionClosed()), SLOT(bs_connectionClosed()));
	connect(d->bs, SIGNAL(delayedCloseFinished()), SLOT(bs_delayedCloseFinished()));
	connect(d->bs, SIGNAL(readyRead()), SLOT(bs_readyRead()));
	connect(d->bs, SIGNAL(bytesWritten(int)), SLOT(bs_bytesWritten(int)));

	connected();

	// if there are existing bytes, take 'em
	d->spare.resize(0);
	if(d->bs->bytesAvailable())
		d->spare = d->bs->read();

	// immediate SSL?
	if(d->conn->useSSL())
		do_starttls();
	else
		do_stream();
}

void ClientStream::cr_error()
{
	reset();
	error(ErrConnection);
}

void ClientStream::bs_connectionClosed()
{
	reset();
	connectionClosed();
}

void ClientStream::bs_delayedCloseFinished()
{
	// we don't care about this (we track all important data ourself)
}

void ClientStream::bs_readyRead()
{
	QByteArray a = d->bs->read();
	incomingRawData(a);
}

void ClientStream::bs_bytesWritten(int bytes)
{
	int plain = bytes;
	if(d->tls_started)
		plain = d->layerTLS.finished(plain);
	if(d->sasl_authed)
		plain = d->layerSASL.finished(plain);
	if(plain > 0)
		dataWritten(plain);
}

void ClientStream::do_starttls()
{
	d->tls_started = true;
	d->tls->startClient(d->server);

	// Calling startClient might cause a signal to come back and reset everything, so don't do the rest
	// unless we are still good
	if(d->bs) {
		QByteArray a = d->spare.copy();
		d->spare.resize(0);
		incomingRawData(a);
	}
}

void ClientStream::tls_success()
{
	securityLayerActivated();
	do_stream();
}

void ClientStream::tls_fail()
{
	reset();
	d->errCond = TLSFail;
	error(ErrTLS);
}

void ClientStream::tls_closed()
{
	reset();
	connectionClosed();
}

void ClientStream::tls_readyRead(const QByteArray &a)
{
	incomingTLSDecryptedData(a);
}

void ClientStream::tls_readyReadOutgoing(const QByteArray &a, int plainBytes)
{
	if(d->tls_started)
		d->layerTLS.specifyEncoded(a.size(), plainBytes);
	writeRawData(a);
}

void ClientStream::sasl_clientFirstStep(const QString &mech, const QByteArray *stepData)
{
	d->sasl_mech = mech;
	QDomElement e = d->doc.createElementNS(NS_SASL, "auth");
	e.setAttribute("mechanism", mech);
	if(stepData && !stepData->isEmpty()) {
#ifdef XMPP_TEST
		QCString cs(stepData->data(), stepData->size()+1);
		TD::msg(QString("SASL OUT: [%1]").arg(QString::fromLatin1(cs)));
#endif
		e.appendChild(d->doc.createTextNode(Base64::arrayToString(*stepData)));
	}
	writeElement(e, true);

	d->state = NeedSASLChallenge;

	// see if the response is ready (highly unlikely, but just in case)
	do_getNextChallenge();
}

void ClientStream::sasl_nextStep(const QByteArray &stepData)
{
#ifdef XMPP_TEST
	QCString cs;
	cs.resize(stepData.size()+1);
	memcpy(cs.data(), stepData.data(), stepData.size());
	TD::msg(QString("SASL OUT: [%1]").arg(QString::fromLatin1(cs)));
#endif
	QDomElement e = d->doc.createElementNS(NS_SASL, "response");
	if(!stepData.isEmpty())
		e.appendChild(d->doc.createTextNode(Base64::arrayToString(stepData)));
	writeElement(e, true);

	d->state = NeedSASLChallenge;

	// see if the response is ready (highly unlikely, but just in case)
	do_getNextChallenge();
}

void ClientStream::sasl_needParams(bool user, bool authzid, bool pass, bool realm)
{
	if(authzid)
		d->sasl->setAuthzid(d->jid.full());
	if(user || pass || realm) {
		d->state = NeedParams;
		needAuthParams(user, pass, realm);
	}
	else
		d->sasl->continueAfterParams();
}

void ClientStream::sasl_authenticated()
{
	d->sasl_ssf = d->sasl->ssf();
}

void ClientStream::sasl_readyRead()
{
	QByteArray a = d->sasl->read();
	incomingData(a);
}

void ClientStream::sasl_readyReadOutgoing(int plainBytes)
{
	QByteArray a = d->sasl->readOutgoing();
	if(d->sasl_authed)
		d->layerSASL.specifyEncoded(a.size(), plainBytes);
	writeSASLEncryptedData(a);
}

void ClientStream::sasl_error(int x)
{
	if(x == QCA::SASL::ErrAuth) {
		int x = convertedSASLCond();
		reset();
		d->errCond = x;
		error(ErrAuth);
	}
	else {
		reset();
		error(ErrSecurityLayer);
	}
}

void ClientStream::incomingRawData(const QByteArray &a)
{
	//QString str = QCA::arrayToHex(a);
	//printf("in[%d]: {%s}\n", a.size(), str.latin1());

	if(d->tls_started)
		d->tls->writeIncoming(a);
	else if(d->sasl_authed)
		d->sasl->writeIncoming(a);
	else
		incomingData(a);
}

void ClientStream::incomingTLSDecryptedData(const QByteArray &a)
{
	if(d->sasl_authed)
		d->sasl->writeIncoming(a);
	else
		incomingData(a);
}

void ClientStream::writeSASLEncryptedData(const QByteArray &a)
{
	if(d->tls_started) {
		d->layerTLS.addPlain(a.size());
		d->tls->write(a);
	}
	else
		writeRawData(a);
}

void ClientStream::writeRawData(const QByteArray &a)
{
	//QString str = QCA::arrayToHex(a);
	//printf("out[%d]: {%s}\n", a.size(), str.latin1());
	d->bs->write(a);
}

void ClientStream::writeData(const QByteArray &a)
{
	if(a.isEmpty())
		return;

#ifdef XMPP_DEBUG
	QCString cs(a.data(), a.size()+1);
	fprintf(stderr, "ClientStream: send: [%s]\n", cs.data());
#endif

	if(d->state != Active) {
		QCString cs(a.size()+1);
		memcpy(cs.data(), a.data(), a.size());
		QString str = QString::fromUtf8(cs);
		outgoingXml(str);
	}

	if(d->sasl_authed) {
		d->layerSASL.addPlain(a.size());
		d->sasl->write(a);
	}
	else if(d->tls_started) {
		d->layerTLS.addPlain(a.size());
		d->tls->write(a);
	}
	else
		writeRawData(a);
}

void ClientStream::dataWritten(int bytes)
{
	//printf("wrote: [%d]\n", bytes);

	if(d->state == Active || d->state == Closing) {
		int numComplete = 0;
		for(QValueList<TrackItem>::Iterator it = d->trackQueue.begin(); it != d->trackQueue.end();) {
			TrackItem &i = *it;

			// enough bytes?
			if(bytes < i.size) {
				i.size -= bytes;
				break;
			}
			int type = i.type;
			bytes -= i.size;
			it = d->trackQueue.remove(it);
			++numComplete;

			if(type == TrackItem::Stanza)
				stanzaWritten();
			else if(type == TrackItem::Close) {
				reset(true);
				delayedCloseFinished();
				return;
			}
		}

		//if(numComplete > 0)
		//	printf("completed %d items\n", numComplete);
	}
}

void ClientStream::incomingData(const QByteArray &a)
{
	// don't accept data if we are shutting down
	if(a.isEmpty() || d->state == Closing)
		return;
#ifdef XMPP_DEBUG
	QCString cs(a.data(), a.size()+1);
	fprintf(stderr, "ClientStream: recv: %d [%s]\n", a.size(), cs.data());
#endif
	d->xml.appendData(a);

	if(d->state == WaitVersion || d->state == WaitTLS || d->state == NeedSASLStep || d->state == NeedParams)
		return;
	processPackets();
}

void ClientStream::processPackets()
{
	QGuardedPtr<ClientStream> self = this;

	while(parseNext());

	if(!self)
		return;
	if(!d->in.isEmpty())
		readyRead();
}

bool ClientStream::parseNext()
{
	Parser::Event e = d->xml.readNext();
	if(e.isNull()) // not yet
		return false;
	int t = e.type();
	if(t == Parser::Event::Error) {
		reset();
		error(ErrParse);
		return false;
	}

	if(d->state != Active)
		incomingXml(e.actualString());

	if(d->state == NeedStream) {
		// first event is always a DocumentOpen
		if(e.namespaceURI() == NS_STREAMS && e.localName() == "stream") {
			QXmlAttributes atts = e.atts();
			d->from = atts.value("from");
			d->id = atts.value("id");
			if(atts.value("version") == "1.0" && !d->oldOnly)
				d->old = false;
			else
				d->old = true;

#ifdef XMPP_TEST
			TD::msg(QString("handshake success (from=[%1], id=[%2])").arg(d->from).arg(d->id));
#endif
			if(d->old) {
				d->state = WaitVersion;
				warning(WarnOldVersion);
				return false;
			}
			do_afterStream();
			return false;
		}
		else {
			reset();
			error(ErrProtocol);
			return false;
		}
	}
	else {
		if(t == Parser::Event::DocumentClose) {
			reset();
			connectionClosed();
			return false;
		}
		else // must be stanza
			return incomingStanza(e.stanza());
	}
}

bool ClientStream::incomingStanza(const QDomElement &e)
{
#ifdef XMPP_DEBUG
	printf("* Stanza [%s]\n", e.tagName().latin1());
	printf("  ns: [%s]\n", e.namespaceURI().latin1());
	printf("  attribs:");
	QDomNamedNodeMap atts = e.attributes();
	for(int n = 0; n < (int)atts.length(); ++n) {
		QDomAttr a = atts.item(n).toAttr();
		QString ns = a.namespaceURI();
		printf(" %s=[%s]", a.name().latin1(), a.value().latin1());
		if(!ns.isNull())
			printf(" (%s)", ns.latin1());
	}
	printf("\n");
#endif

	if(d->state == Active) {
		Stanza s = createStanza(e);
		if(s.isNull())
			return true;
		d->in.append(new Stanza(s));
		return true;
	}

#ifdef XMPP_TEST
	TD::incomingXml(e);
#endif

	if(d->old) {
		if(d->state == NeedAuthGetResponse) {
			// waiting for an iq
			if(e.namespaceURI() == NS_CLIENT && e.tagName() == "iq") {
				Jid from(e.attribute("from"));
				QString type(e.attribute("type"));
				QString id(e.attribute("id"));

				bool okfrom = (from.isEmpty() || from.compare(Jid(d->server)));
				if(okfrom && id == "auth_1" && (type == "result" || type == "error")) {
					if(type == "result") {
						QDomElement q = e.elementsByTagNameNS("jabber:iq:auth", "query").item(0).toElement();
						if(q.isNull() || q.elementsByTagName("username").item(0).isNull() || q.elementsByTagName("resource").item(0).isNull()) {
							reset();
							error(ErrProtocol);
							return false;
						}
						bool plain = !q.elementsByTagName("password").item(0).isNull();
						bool digest = !q.elementsByTagName("digest").item(0).isNull();

						if(!digest && !plain) {
							reset();
							error(ErrProtocol);
							return false;
						}

						// plain text not allowed?
						if(!digest && !d->allowPlain) {
							reset();
							d->errCond = NoMech;
							error(ErrAuth);
							return false;
						}

						d->state = NeedParams;
						d->digest = digest;
						needAuthParams(false, true, false);
					}
					else {
						int x = Private::getOldErrorCode(e);
						if(x != -1)
							x = Private::authErrorOldToNew(x);
						if(x == -1)
							x = GenericAuthError;
						reset();
						d->errCond = x;
						error(ErrAuth);
						return false;
					}
				}
				else {
					// ignore
				}
			}
			else {
				// ignore
			}
		}
		else if(d->state == NeedAuthSetResponse) {
			// waiting for an iq
			if(e.namespaceURI() == NS_CLIENT && e.tagName() == "iq") {
				Jid from(e.attribute("from"));
				QString type(e.attribute("type"));
				QString id(e.attribute("id"));

				bool okfrom = (from.isEmpty() || from.compare(Jid(d->server)));
				if(okfrom && id == "auth_2" && (type == "result" || type == "error")) {
					if(type == "result") {
						d->state = Active;
						setNoopTime(d->noop_time);
						authenticated();
					}
					else {
						int x = Private::getOldErrorCode(e);
						if(x != -1)
							x = Private::authErrorOldToNew(x);
						if(x == -1)
							x = GenericAuthError;
						reset();
						d->errCond = x;
						error(ErrAuth);
						return false;
					}
				}
				else {
					// ignore
				}
			}
			else {
				// ignore
			}
		}
	}
	else {
		// error stanza?
		if(e.namespaceURI() == NS_STREAMS && e.tagName() == "error") {
			int x;
			QString text;
			QDomElement appSpec;

			QDomElement t = e.firstChild().toElement();
			if(t.isNull() || t.namespaceURI() != NS_STREAMS)
				x = -1;
			else
				x = Private::stringToStreamCond(t.tagName());
			t = e.elementsByTagNameNS(NS_STREAMS, "text").item(0).toElement();
			if(!t.isNull())
				text = t.text();

			// find first non-standard namespaced element
			QDomNodeList nl = e.childNodes();
			for(uint n = 0; n < nl.count(); ++n) {
				QDomNode i = nl.item(n);
				if(i.isElement() && i.namespaceURI() != NS_STREAMS) {
					appSpec = i.toElement();
					break;
				}
			}

			int connErr = -1;
			int strErr = -1;
			if(x == Private::BadFormat) {} // should NOT happen (we send the right format)
			else if(x == Private::BadNamespacePrefix) {} // should NOT happen (we send prefixes)
			else if(x == Private::Conflict)
				strErr = Conflict;
			else if(x == Private::ConnectionTimeout)
				strErr = ConnectionTimeout;
			else if(x == Private::HostGone)
				connErr = HostGone;
			else if(x == Private::HostUnknown)
				connErr = HostUnknown;
			else if(x == Private::ImproperAddressing) {} // should NOT happen (we aren't a server)
			else if(x == Private::InternalServerError)
				strErr = InternalServerError;
			else if(x == Private::InvalidId) {} // should NOT happen (clients don't specify id)
			else if(x == Private::InvalidNamespace) {} // should NOT happen (we set the right ns)
			else if(x == Private::InvalidXml)
				strErr = InvalidXml; // shouldn't happen either, but just in case ...
			else if(x == Private::NonmatchingHosts) {} // should NOT happen (we aren't a server)
			else if(x == Private::StreamNotAuthorized) {} // should NOT happen (we're not stupid)
			else if(x == Private::PolicyViolation)
				strErr = PolicyViolation;
			else if(x == Private::RemoteConnectionFailed)
				connErr = RemoteConnectionFailed;
			else if(x == Private::ResourceConstraint)
				strErr = ResourceConstraint;
			else if(x == Private::RestrictedXml)
				strErr = InvalidXml; // group with this one
			else if(x == Private::SeeOtherHost)
				connErr = SeeOtherHost;
			else if(x == Private::SystemShutdown)
				strErr = SystemShutdown;
			else if(x == Private::UndefinedCondition) {} // leave as null error
			else if(x == Private::UnsupportedEncoding) {} // should NOT happen (we send good encoding)
			else if(x == Private::UnsupportedStanzaType) {} // should NOT happen (we're not stupid)
			else if(x == Private::UnsupportedVersion)
				connErr = UnsupportedVersion;
			else if(x == Private::XmlNotWellFormed)
				strErr = InvalidXml; // group with this one
			reset();
			d->errText = text;
			d->errAppSpec = appSpec;
			if(connErr != -1) {
				d->errCond = connErr;
				error(ErrNeg);
			}
			else {
				if(strErr != -1)
					d->errCond = strErr;
				else
					d->errCond = GenericStreamError;
				error(ErrStream);
			}
			return false;
		}

		if(d->state == NeedFeatures) {
			// we are waiting for stream features
			if(e.namespaceURI() == NS_STREAMS && e.tagName() == "features") {
				// extract features
				StreamFeatures f;
				QDomElement s = e.elementsByTagNameNS(NS_TLS, "starttls").item(0).toElement();
				if(!s.isNull()) {
					f.tls_supported = true;
					f.tls_required = s.elementsByTagNameNS(NS_TLS, "required").count() > 0;
				}
				QDomElement m = e.elementsByTagNameNS(NS_SASL, "mechanisms").item(0).toElement();
				if(!m.isNull()) {
					f.sasl_supported = true;
					QDomNodeList l = m.elementsByTagNameNS(NS_SASL, "mechanism");
					for(uint n = 0; n < l.count(); ++n)
						f.sasl_mechs += l.item(n).toElement().text();
				}

				if(f.tls_supported) {
#ifdef XMPP_TEST
					QString s = "STARTTLS is available";
					if(f.tls_required)
						s += " (required)";
					TD::msg(s);
#endif
				}
				if(f.sasl_supported) {
#ifdef XMPP_TEST
					QString s = "SASL mechs:";
					for(QStringList::ConstIterator it = f.sasl_mechs.begin(); it != f.sasl_mechs.end(); ++it)
						s += QString(" [%1]").arg((*it));
					TD::msg(s);
#endif
				}
				d->features = f;
				do_handleFeatures();
				return false;
			}
			else {
				// ignore
			}
		}
		else if(d->state == NeedTLSProceed) {
			// waiting for proceed to starttls
			if(e.namespaceURI() == NS_TLS) {
				if(e.tagName() == "proceed") {
#ifdef XMPP_TEST
					TD::msg("Server wants us to proceed with ssl handshake");
#endif
					d->spare = d->xml.unprocessed();
					d->xml.reset();
					do_starttls();
					return false;
				}
				else if(e.tagName() == "failure") {
					reset();
					d->errCond = TLSStart;
					error(ErrTLS);
					return false;
				}
				else {
					reset();
					error(ErrProtocol);
					return false;
				}
			}
			else {
				// ignore
			}
		}
		else if(d->state == NeedSASLChallenge) {
			// waiting for sasl challenge/success/fail
			if(e.namespaceURI() == NS_SASL) {
				if(e.tagName() == "challenge") {
					QByteArray a = Base64::stringToArray(e.text());
#ifdef XMPP_TEST
					QCString cs;
					cs.resize(a.size()+1);
					memcpy(cs.data(), a.data(), a.size());
					TD::msg(QString("SASL IN: [%1]").arg(QString::fromLatin1(cs)));
#endif
					d->state = NeedSASLStep;
					d->sasl->putStep(a);
					return false;
				}
				else if(e.tagName() == "success") {
					d->spare = d->xml.unprocessed();
					d->sasl_authed = true;
					d->xml.reset();
					if(d->sasl_ssf > 0)
						securityLayerActivated();
					do_stream();
					return false;
				}
				else if(e.tagName() == "failure") {
					QDomElement t = e.firstChild().toElement();
					int x;
					if(t.isNull() || t.namespaceURI() != NS_SASL)
						x = GenericAuthError;
					x = Private::stringToSASLCond(t.tagName());
					if(x == Private::Aborted) // should NOT happen (we never send <abort/>)
						x = GenericAuthError;
					else if(x == Private::BadProtocol) // should NOT happen (we obey the protocol)
						x = GenericAuthError;
					else if(x == Private::EncryptionRequired)
						x = EncryptionRequired;
					else if(x == Private::InvalidAuthzid)
						x = InvalidAuthzid;
					else if(x == Private::InvalidMech)
						x = InvalidMech;
					else if(x == Private::InvalidRealm)
						x = InvalidRealm;
					else if(x == Private::MechTooWeak)
						x = MechTooWeak;
					else if(x == Private::NotAuthorized)
						x = NotAuthorized;
					else if(x == Private::TemporaryAuthFailure)
						x = TemporaryAuthFailure;
					else
						x = GenericAuthError;
					reset();
					d->errCond = x;
					error(ErrAuth);
					return false;
				}
				else {
					reset();
					error(ErrProtocol);
					return false;
				}
			}
			else {
				// ignore
			}
		}
	}

	return true;
}

void ClientStream::do_stream()
{
#ifdef XMPP_TEST
	TD::msg("Starting root stream...");
#endif
	QString xmlHeader;
	createStreamXmlTags(NS_CLIENT, d->server, d->oldOnly ? QString() : QString("1.0"), &xmlHeader, &d->streamOpen, &d->streamClose);

	QString s;
	s += xmlHeader + '\n';
	s += d->streamOpen + '\n';

#ifdef XMPP_TEST
	TD::outgoingTag(xmlHeader);
	TD::outgoingTag(d->streamOpen);
#endif

	d->state = NeedStream;
	writeString(s);

	// try spare data (triggers a 'processPackets')
	QByteArray a = d->spare.copy();
	d->spare.resize(0);
	incomingRawData(a);
}

void ClientStream::do_afterStream()
{
	if(d->old) {
		// In the old protocol, TLS must happen before the stream.  So if we're here and we don't
		// have TLS yet, then we're never going to have it.
		if(!d->tls_started && !d->tls_warned) {
			d->tls_warned = true;
			d->state = WaitTLS;
			warning(WarnNoTLS);
			return;
		}

		if(!d->doAuth) {
			d->state = Active;
			setNoopTime(d->noop_time);
			authenticated();
			return;
		}

		d->state = NeedAuthGetResponse;
		QDomElement e = d->doc.createElement("iq");
		e.setAttribute("to", d->server);
		e.setAttribute("type", "get");
		e.setAttribute("id", "auth_1");
		QDomElement q = d->doc.createElementNS("jabber:iq:auth", "query");
		QDomElement u = d->doc.createElement("username");
		u.appendChild(d->doc.createTextNode(d->jid.node()));
		q.appendChild(u);
		e.appendChild(q);
		writeElement(e);
	}
	else {
		d->state = NeedFeatures;
		processPackets();
	}
}

void ClientStream::do_handleFeatures()
{
	// deal with TLS?
	if(!d->tls_started && !d->sasl_authed) {
		if(d->features.tls_supported) {
			d->state = NeedTLSProceed;
			QDomElement e = d->doc.createElementNS(NS_TLS, "starttls");
			writeElement(e, true);
			return;
		}
		else if(!d->tls_warned) {
			d->tls_warned = true;
			d->state = WaitTLS;
			warning(WarnNoTLS);
			return;
		}
	}

	// deal with SASL?
	if(!d->sasl_authed) {
		if(!d->features.sasl_supported) {
			reset();
			error(ErrProtocol);
			return;
		}

#ifdef XMPP_TEST
		TD::msg("starting SASL authentication...");
#endif

		// no SASL plugin?  fall back to Simple SASL
		if(!QCA::isSupported(QCA::CAP_SASL)) {
			// Simple SASL needs MD5.  do we have that either?
			if(!QCA::isSupported(QCA::CAP_MD5))
				QCA::insertProvider(createProviderHash());
			QCA::insertProvider(createProviderSimpleSASL());
		}

		d->sasl = new QCA::SASL;
		connect(d->sasl, SIGNAL(clientFirstStep(const QString &, const QByteArray *)), SLOT(sasl_clientFirstStep(const QString &, const QByteArray *)));
		connect(d->sasl, SIGNAL(nextStep(const QByteArray &)), SLOT(sasl_nextStep(const QByteArray &)));
		connect(d->sasl, SIGNAL(needParams(bool, bool, bool, bool)), SLOT(sasl_needParams(bool, bool, bool, bool)));
		connect(d->sasl, SIGNAL(authenticated()), SLOT(sasl_authenticated()));
		connect(d->sasl, SIGNAL(readyRead()), SLOT(sasl_readyRead()));
		connect(d->sasl, SIGNAL(readyReadOutgoing(int)), SLOT(sasl_readyReadOutgoing(int)));
		connect(d->sasl, SIGNAL(error(int)), SLOT(sasl_error(int)));

		if(d->haveLocalAddr)
			d->sasl->setLocalAddr(d->localAddr, d->localPort);
		if(d->conn->havePeerAddress())
			d->sasl->setRemoteAddr(d->conn->peerAddress(), d->conn->peerPort());

		d->sasl->setAllowPlain(d->allowPlain);
		d->sasl->setRequireMutualAuth(d->mutualAuth);

		// only use SASL security layer if we don't have TLS
		if(!d->tls_started) {
			d->sasl->setMinimumSSF(d->minimumSSF);
			d->sasl->setMaximumSSF(d->maximumSSF);
		}

		d->state = NeedSASLStep;
		QStringList ml;
		if(!d->sasl_mech.isEmpty())
			ml += d->sasl_mech;
		else
			ml = d->features.sasl_mechs;
		if(!d->sasl->startClient("xmpp", d->server, ml, false)) {
			int x = convertedSASLCond();
			reset();
			d->errCond = x;
			error(ErrAuth);
			return;
		}
		return;
	}

	d->state = Active;
	setNoopTime(d->noop_time);
	authenticated();
}

void ClientStream::do_getNextChallenge()
{
	processPackets();
}

void ClientStream::do_authset()
{
	d->state = NeedAuthSetResponse;
	QDomElement e = d->doc.createElement("iq");
	e.setAttribute("to", d->server);
	e.setAttribute("type", "set");
	e.setAttribute("id", "auth_2");
	QDomElement q = d->doc.createElementNS("jabber:iq:auth", "query");
	QDomElement u = d->doc.createElement("username");
	u.appendChild(d->doc.createTextNode(d->jid.node()));
	q.appendChild(u);
	QDomElement p;
	if(d->digest) {
		// need SHA1 here
		if(!QCA::isSupported(QCA::CAP_SHA1))
			QCA::insertProvider(createProviderHash());

		p = d->doc.createElement("digest");
		QCString cs = d->id.utf8() + d->pass.utf8();
		p.appendChild(d->doc.createTextNode(QCA::SHA1::hashToString(cs)));
	}
	else {
		p = d->doc.createElement("password");
		p.appendChild(d->doc.createTextNode(d->pass));
	}
	q.appendChild(p);
	QDomElement r = d->doc.createElement("resource");
	r.appendChild(d->doc.createTextNode(d->jid.resource()));
	q.appendChild(r);
	e.appendChild(q);
	writeElement(e);
}

int ClientStream::writeElement(const QDomElement &e, bool clip)
{
#ifdef XMPP_TEST
	if(d->state != Active)
		TD::outgoingXml(e);
#endif
	return Stream::writeElement(e, clip);
}

int ClientStream::convertedSASLCond() const
{
	int x = d->sasl->errorCondition();
	if(x == QCA::SASL::NoMech)
		return NoMech;
	else if(x == QCA::SASL::BadProto)
		return BadProto;
	else if(x == QCA::SASL::BadServ)
		return BadServ;
	else if(x == QCA::SASL::TooWeak)
		return MechTooWeak;
	else
		return GenericAuthError;
}

void ClientStream::doNoop()
{
	if(d->state == Active) {
		TrackItem i;
		i.type = TrackItem::Ping;
		i.size = writeString("\n");
		d->trackQueue += i;
	}
}

//----------------------------------------------------------------------------
// Debug
//----------------------------------------------------------------------------
Debug::~Debug()
{
}

TD::TD()
{
}

TD::~TD()
{
}

void TD::msg(const QString &s)
{
	if(debug_ptr)
		debug_ptr->msg(s);
}

void TD::outgoingTag(const QString &s)
{
	if(debug_ptr)
		debug_ptr->outgoingTag(s);
}

void TD::incomingTag(const QString &s)
{
	if(debug_ptr)
		debug_ptr->incomingTag(s);
}

void TD::outgoingXml(const QDomElement &e)
{
	if(debug_ptr)
		debug_ptr->outgoingXml(e);
}

void TD::incomingXml(const QDomElement &e)
{
	if(debug_ptr)
		debug_ptr->incomingXml(e);
}
